import logging

from stl import mesh, Mode
from sect.triangulation import Triangulation
from ground.base import get_context

from rocolib.api.composables.graph.HyperEdge import HyperEdge
from rocolib.utils.utils import prefix as prefixString
import rocolib.utils.numsym as np


log = logging.getLogger(__name__)

def inflate(face, thickness=.1, edges=False):
    dt = np.array([[0],[0],[thickness/2.],[0]])
    nf = face-dt
    pf = face+dt

    faces = []

    if edges:
        faces.append(np.transpose(np.array((pf[:,0], nf[:,0], pf[:,1]))))
        faces.append(np.transpose(np.array((nf[:,0], nf[:,1], pf[:,1]))))
    else:
        faces.append(pf)          # top face
        faces.append(nf[:,::-1])  # bottom face

    return faces


def _triangulate(faces, thickness=0):
    """Create triangulated faces in 3D space with thickness"""
    scale = .001  # roco units : mm ; STL units m

    tris = []

    nparr = lambda l: np.transpose([list(x) + [0, 1] for x in l])

    for f in faces:
        if f.area == 0:
            log.debug(f"Omitting face {f.name} with area {f.area} from solid model")
            continue # skip faces with 0 area
        r = f.transform3D
        facets_coord_index = []
        facets_normal = []
        facets = []

        def add_triangles(t, edges=False):
            if thickness:
                facets = inflate(t, thickness=thickness, edges=edges)
            elif edges:
                return []
            else:
                facets = [t]

            for x in facets:
                tri = np.transpose(np.dot(r, x)[0:3] * scale).tolist()
                tris.append(tri)


        poly = f.polygon()
        triangles = ( x.vertices for x in Triangulation.constrained_delaunay(poly, context=get_context()).triangles() )

        for t in map(nparr, triangles):
            add_triangles(t)
        for t in map(nparr, poly.edges):
            add_triangles(t, edges=True)

    data = np.zeros(len(tris), dtype=mesh.Mesh.dtype)
    solid = mesh.Mesh(data)
    for i, tri in enumerate(tris):
        for j in range(3):
            solid.vectors[i][j] = tri[j]
    return solid


class Graph():
    def __init__(self):
        self.faces = []
        self.facelists = []
        self.edges = []
        self._thickness = 0
        self._meshes = []
        self._fullmesh = None

    @property
    def meshes(self):
        if not self._meshes:
            self.makeMeshes()
        return self._meshes

    @property
    def fullmesh(self):
        if not self._fullmesh:
            self.makeMeshes()
        return self._fullmesh

    def addFace(self, f, prefix=None, root=False, faceEdges=None, faceAngles=None, faceFlips=None):
        if prefix:
            f.prefix(prefix)
        if f in self.faces:
            raise ValueError("Face %s already in graph" % f.name)
        if root:
            self.faces.insert(0, f)
        else:
            self.faces.append(f)

        if faceEdges is not None:
            f.renameEdges(faceEdges, faceAngles, faceFlips, self.edges)
            if prefix:
                f.prefixEdges(prefix)

        self.rebuildEdges()
        return self

    def attachFace(self, fromFace, toFace, prefix=None, transform=None):
        self.addFace(toFace, prefix)
        self.mergeFace(fromFace, toFace.name, transform)

    def mergeFace(self, fromFaceName, toFaceName, transform=None, offset=(0,0), copyDecorations=False):
        fromFace = self.getFace(fromFaceName)
        toFace = self.getFace(toFaceName)
        toFace.transform(origin=offset)
        if transform is None:
            transform = np.eye(4)
            invtransform = np.eye(4)
        elif transform is False:
            transform = False
            invtransform = False
        else:
            invtransform = np.inv(transform)

        fromFace.addFace(toFace, transform, copyDecorations)
        toFace.addFace(fromFace, invtransform, copyDecorations=False)
        # XXX TODO have separate options for copyDecorations from->to vs to->from?
        # Right now we'll do both during place, so only note it for one of the pair.

    def attachEdge(self, fromEdge, newFace, newEdge, prefix=None, root=False, angle=0, edgeType=None, joints=None):
        # XXX should set angle from a face, not absolute angle of the face
        self.addFace(newFace, prefix, root)

        if fromEdge is not None:
            newEdge = prefixString(prefix, newEdge)
            self.mergeEdge(fromEdge, newEdge, angle=angle, edgeType=edgeType, joints=joints)

    def delFace(self, facename):
        for (i, f) in enumerate(self.faces):
            if f.name == facename:
                f.disconnectAll()
                self.faces.pop(i)
                self.rebuildEdges()
                return self

        return self

    def getFace(self, name):
        for f in self.faces:
            if f.name == name:
                return f
        return None

    def getEdge(self, name):
        for e in self.edges:
            if e.name == name:
                return e
        return None

    def prefix(self, prefix):
        for e in self.edges:
            e.rename(prefixString(prefix, e.name))
        for f in self.faces:
            f.rename(prefixString(prefix, f.name))

    def renameEdge(self, fromname, toname):
        e = self.getEdge(fromname)
        if e:
            e.rename(toname)

    def rebuildEdges(self):
        self.edges = []
        for f in self.faces:
            for e in f.edges:
                if e not in self.edges:
                    self.edges.append(e)

    def invertEdges(self):
        # swap mountain and valley folds
        for e in self.edges:
            for f in e.faces:
                e.faces[f] = (-e.faces[f][0], e.faces[f][1])
        for f in self.faces:
            f.inverted = not f.inverted

    def addTab(self, edge1, edge2, angle=0, width=10):
        self.mergeEdge(edge1, edge2, angle=angle, tabWidth=width)

    def mergeEdge(self, edge1, edge2, angle=0, tabWidth=None, edgeType=None, joints=None, swap=False):
        e1 = self.getEdge(edge1)
        e2 = self.getEdge(edge2)
        if e1 is None or e2 is None:
            logstr  = f"Edge not found trying to merge ({edge1}, {edge2}) in edges:\n  "
            logstr += "\n  ".join([e.name for e in self.edges])
            log.error(logstr)
            raise AttributeError("Edge not found")
        if swap:
            e1, e2 = e2, e1

        if len(e2.faces) > 1:
            log.warning("Adding more than two faces to an edge, currently not well supported.")
            e2.mergeWith(e1, angle=angle, flip=False, tabWidth=tabWidth)
        else:
            e2.mergeWith(e1, angle=angle, flip=True, tabWidth=tabWidth)
        self.edges.remove(e1)
        
        e2.setType(edgeType)
        if joints:
            for joint in joints.joints:
                e2.addJoint(joint)

        return self

    def splitEdge(self, edge):
        old_edge = edge
        old_edge_name = edge.name
        new_edges_and_faces = []

        for i, face in enumerate(list(old_edge.faces)):
            length = old_edge.length
            angle = old_edge.faces[face][0]
            flip = old_edge.faces[face][1]

            new_edge_name = old_edge_name + '.se' + str(i)
            new_edge = HyperEdge(new_edge_name, length)
            face.replaceEdge(old_edge, new_edge, angle, flip=False )
            new_edges_and_faces.append((new_edge_name, face, length, angle, flip))

        self.rebuildEdges()
        return new_edges_and_faces

    def jointify(self, **kwargs):
        for e in self.edges:
            if e.isNotFlat() and "joint" in kwargs:
                e.setType("JOINT")
                e.addJoint(kwargs["joint"])
                #print "jointing ", e.name

    def tabify(self, tabFace=None, tabDecoration=None, slotFace=None, slotDecoration=None, **kwargs):
        for e in self.edges:
            if e.isTab():
                #print "tabbing ", e.name
                for (edgename, face, length, angle, flip) in self.splitEdge(e):
                    if flip:
                        #print "-- tab on: ", edgename, face.name, angle
                        if tabDecoration is not None and ((abs(angle) > 179.5 and abs(angle) < 180.5) or tabFace is None):
                            tabDecoration(face, edgename, e.tabWidth, flip=True, **kwargs)
                        elif tabFace is not None:
                            tab = tabFace(length, e.tabWidth, **kwargs)
                            self.attachEdge(edgename, tab, tab.MAINEDGE, prefix=edgename, angle=0)
                    else:
                        #print "-- slot on: ", edgename, face.name
                        if slotFace is not None:
                            # XXX TODO: set angle appropriately
                            slot = slotFace(length, e.tabWidth, **kwargs)
                            self.attachEdge(edgename, slot, slot.MAINEDGE, prefix=edgename, angle=0)
                        if slotDecoration is not None:
                            slotDecoration(face, edgename, e.tabWidth, **kwargs)

        #TODO: extend this to three+ edges
        #component.addConnectors((conn, cname), new_edges[0], new_edges[1], depth, tabattachment=None, angle=0)

    def flip(self):
        raise NotImplementedError
        for f in self.faces:
            f.flip()

    def transform(self, scale=1, angle=0, origin=(0,0)):
        raise NotImplementedError

    def dotransform(self, scale=1, angle=0, origin=(0,0)):
        for f in self.faces:
            f.transform(scale, angle, origin)

    def mirror(self):
        for f in self.faces:
            f.mirror()

    def mirrorY(self):
        raise NotImplementedError
        for f in self.faces:
            f.transform( mirrorY())

    def mirrorX(self):
        raise NotImplementedError
        for f in self.faces:
            f.transform( mirrorX())

    def toString(self):
        print()
        for f in self.faces:
            print(f.name + repr(f.edges))

    def graphObj(self):
        g = {}
        for f in self.faces:
            g[f.name] = dict([(e and e.name or "", e) for e in f.edges])
        return g

    def showGraph(self):
        import objgraph
        objgraph.show_refs(self.graphObj(), max_depth = 2, filter = lambda x : isinstance(x, (dict, HyperEdge)))

    def place(self, transform3D=None):
        if transform3D is None:
            transform3D = np.eye(4)

        while True:
            for f in self.faces:
                if f.transform2D is not None and f.transform3D is not None:
                    continue
                else:
                    f.place(None, None, transform3D, self.facelists)
                    break
            else:
                break

        for f in self.faces:
            f.copyDecorations()
        self.rebuildEdges()

    def makeMeshes(self, **kwargs):
        """Create a list of stl Mesh objects"""
        self.place()
        # Create 1 mesh per connected set of faces.
        self._meshes = [_triangulate(faces, self._thickness) for faces in self.facelists]
        self._fullmesh = mesh.Mesh(np.concatenate([m.data for m in self._meshes]))

    def getEdgeSet(self):
        self.place()
        edgeset = {}

        # Tile faces along x axis with buffer spacing between them
        maxx = 0
        buffer = 10

        for flist in self.facelists:
            edges = set((e for face in flist for e in face.edges if e.pts2D is not None))

            if not edges:
                log.warning(f"No placed edges in flist, moving on.")
                continue

            pts = [p for e in edges for p in e.pts2D]
            minx = min([x[0] for x in pts])
            dx = maxx - minx
            dy = 0

            maxx = max([x[0] for x in pts]) + dx + buffer

            for e in edges:
                if len(e.faces) == 2:
                    angles = list(e.faces.values())
                    if angles[0][1]:
                        angle = angles[0][0] - angles[1][0]
                    else:
                        angle = angles[1][0] - angles[0][0]
                else:
                    angle = None

                edge = dict(p0 = e.pts2D[0] + [dx, dy],
                            p1 = e.pts2D[1] + [dx, dy],
                            faces = len(e.faces),
                            angle = angle,
                            length = e.length,
                            edgeType = e.edgeType,
                            interior = False,
                        )
                edgeset[e.name] = edge

            for face in flist:
                for e in face.get2DDecorations():
                    edge = dict(p0 = e[1] + [dx, dy],
                                p1 = e[2] + [dx, dy],
                                faces = None,
                                angle = None,
                                edgeType = e[3],
                                interior = True,
                            )
                    edgeset[e[0]] = edge

        return edgeset
